__code__='GatU0D/[on&>$QoP(Q(,iO.#`00r*.mXS,k&415pn00k_F9eU)>$6kuPF2,]Zb)Q/V!n\\dId)@I?gGqiL\'4:rdC\\6)H?WtmB4a:?6LgPCq(kb:%gflO6]s8Y!t\'8`n.+CNT`?B`7>o8pTIPIKWdltB3a`\'Z&E"8\\1gh!d0]3PYDZGh"*Wqo54:X=YGQMgon0Hp@^sdgH!7(_<"8ir(4Q$CYmf;MWjk/lK@m0b]4T&\'-@1@s6:^q$0C7sk]XpX15hlusW#-S;<1<UgQ\\-R`<\'oKg7\'!#FY?\\RXG*8__Bj^On(.)bi]L/mFJ>-&<:8^rXNAbF*6%KUnX:;*c]"k]SIAa`%&FMV##SZ8Hl=+T^(WCKPqGl/h,\\]bY3jtKAV_>#PZVnM7d%pKHR&!>L@2R$6:3isJ-Yu,Dj\'YCI?>`Vmf?ZRUb\\r/>K9RC;./0h5B/IpD>F,bj5e\\pb2qI9>FHh""od:&s*lATQoS\'H71NNo&A3qFn4B4d%G78qEf):?V^H->rG@&/2-2VlP)`Gs,*XO\'TReXnIf^Qq:&]Ih$q+^Aiu$jL=3T!GQ*<*M"XP:PLJJW,NrA/p:Q0a_KhM,%_q]"UsYe(X.pVmE_9e(/Cm%bB?B3bljnF3cXZH7j%>nMmaMR`JH$29@%:*lBhh4m7o7ag@A\'HA3[2VHT$:>8$b+<S,<?<aX9]2OA.J8`8>q)s1na<,+k+p4L2eB.b25W\'L+d+HM#kgi@3D)c5rd.^_H5fenAFfiX$!JPkgiI(<mNddu%LRl>$9\\bY:Q4,\\<@UDIP9!;!+F<B:HGSESC>?#u8Hd>6+5!])j?aCJ[hZ-6Ks4Ysue\\uEQK.+p")Sf`8O%^3O!ndZI0NotO,\\`3X.9C,\'L#>qS,E:D(.pF2nB`OGTiq&QDQX<Y>-?Uebr\'\'kto(LXrC(-dnJ/9(F"(=]q@_->LsW3GiQ5j_sM42RI6XLa_5f"Dj:\\F7/10oSHbqgCYM)Qq41?X;4lG+.YDYnu\'&k%Jg=M\\7sm4]QYLh[e_6#XI\'XR#pW,O]<Xdk\\N8ih07Z;s)6H"_!gIRj8h>#hn//<d?k]uU@B?*Z^!PuW;I#YoR;/~>'



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/ers/doclet.py
__version__='3.3.0'

"""Doclets allow chunks of content to be reused acrass projects.


A doclet is a smart object which can render itself in the PDF
and HTML worlds. 

"""
# a trivial parser used for completing the doclets
# in an HTML context
try:
    from html.parser import HTMLParser
except:
    from sgmllib import SGMLParser
    HTMLParser = SGMLParser

# used in making the example doclet.
from reportlab.platypus import Paragraph
from reportlab.platypus.tables import Table, TableStyle
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib import colors

from rlextra.radxml.xmlutils import TagWrapper
from xml.sax.saxutils import escape
from rlextra.radxml.xmlmap import MapNode, MapController
class Doclet:
    def setData(self, data):
        """Called by dynamic RML or preprocessor, data hook.

        Doclets can be parameterised and fetch data and this gives them the
        chance to do it."""        
        pass
    
    def asFlowable(self):
        "Return a flowable or list of flowables"

        #the base class does enough to say 'here is a doclet...'
        sty = ParagraphStyle(
                                name='obvious',
                                fontName='Helvetica',
                                fontSize=10,
                                leading=12,
                                backColor=colors.yellow
                                )
        
        return Paragraph('Doclet class: %s' % self.__class__.__name__, style=sty)


    def asHTML(self, look=None):
        return '<p>This is a Doclet of class %s</p>' % self.__class__.__name__

##  To be discussed:
##    def getPdfStyles(self, look=None):
##         "Return a dictionary mapping style names to styles which are used"
##         return {}
##
##    def htmlNeedsExtraFiles(self):
##        "Does it need to create/reference any bitmap files?"
##        return 0
##
##    def getHtmlStyles(self, look=None):
##         "Return a dictionary mapping style names to CSS styles used"
##         return {}






class TestReverseDoclet(Doclet):
    """Reverses the string it was given and displays as table"""

    defaultParaStyle = ParagraphStyle(
                            name='obvious',
                            fontName='Helvetica',
                            fontSize=10,
                            leading=12,
                            backColor=colors.yellow
                            )

    defaultTableStyle = TableStyle([
         ('GRID',(0,0),(-1,-1),2,colors.grey),
         ('BACKGROUND',(0,0),(-1,0),colors.palegreen),
         ('SPAN',(0,0),(-1,0)),
         ('FONT',(0,0),(-1,-1), 'Helvetica',10,12),
         ('FONT',(1,1),(-1,-1), 'Helvetica-Bold',10,12)
         ])

    def __init__(self):
        #may not be needed but let's initialize with dummy data
        # so it can be tested as it.
        self.text = 'hello world'
        self.reversed = 'dlrow olleh'

                  
    def setData(self, data):
        self.text = data

        #reverse it
        l = list(self.text)
        l.reverse()

        self.reversed = ''.join(l)

    def asHTML(self, look=None):
        return """
        <table border>
            <tr><th colspan="2" style="background:pink">Reverse Doclet Example</th></tr>
            <tr><td>Normal:</td><td>%s</td></tr>
            <tr><td>Reversed:</td><td>%s</td></tr>
        </table>
        """ % (self.text, self.reversed)

    def asFlowable(self):
    
        tableData = [
            ["Reverse Doclet Example",""],
            ["Normal:", self.text],
            ["Reversed:", self.reversed]
            ]
        t = Table(tableData, style=self.defaultTableStyle, colWidths=(100,100))
        
        return t        


##class InspectorDoclet(Doclet):
##    """Little table showing info about the data object"""
##
##    def setData(self, data):
##        self.data = data
##
##        self.typeStr = data.__class__.__name__
##        
##        self.attrs = attrs
##
##    def asHTML(self, look=None):
##        out = ["<table border>\n" +
##        '    <tr><th colspan="2" style="background:pink">%s</th></tr>' % self.title]
##        if self.attrs:
##            for (key, value) in self.attrs:
##                out.append("\n    <tr><td>%s</td><td><b>%s</b></td></tr>" % (key, value))
##        else:
##            out.append('\n    <tr><td colspan="2"><i>No attributes supplied</i></td></tr>')
##        out.append("\n</table>")
##        return ''.join(out)
        


class DocletToHtmlMapNode(MapNode):
    """This is a MapNode which expands doclets to HTML"""

    def __init__(self):
         MapNode.__init__(self, None, None, None)

    def translate(self, nodetuple, controller, overrides, containerdict=None):
        (tagName, attrs, children, stuff) = nodetuple
        if '__doclet__' in attrs:
            theDoclet = attrs['__doclet__']
            return theDoclet.asHTML()
        else:
            return ''

class DocletToHtmlMapper(MapController):
    """Top level map controller which expands doclet tags"""
    def __init__(self, topLevelSubstitution = "%s", naive=0, eoCB=None):
        MapController.__init__(self, topLevelSubstitution, naive, eoCB)
        self['doclet'] = DocletToHtmlMapNode()

        



def _makeXMLTag(name,attrs,contents,inner=1):
    if name == 'doclet':
        return attrs['__doclet__'].asHTML()
    else:
        T = []
        a = T.append
        if inner:
            a('\n<%s' % name)
            for k,v in list(attrs.items()):
                a('%s="%s"' % (k,escape(v)))
            T[:] = [' '.join(T)]
        if contents in ([],None):
            if inner: a('/>')
        else:
            if inner: a('>')
            for c in contents:
                a(createHTML(c,inner=1))
            if inner: a('</%s>' % name)
        return ''.join(T)

def createHTML(tag,inner=0):
    '''Takes a TagWrapper and converts to XML,
    if inner=1 then the outer tag is made explicit'''
    R = []
    r = R.append
    if isinstance(tag,(list,tuple)):
        r(_makeXMLTag(tag[0],tag[1] or {},tag[2]))
    elif isinstance(tag,TagWrapper):
        r(_makeXMLTag(tag.tagName,tag._attrs,tag._children,inner=inner))
    else:
        #r(escape(tag))
        r(tag)
    return ''.join(R)




class DocletParser(HTMLParser):
    """Passes through an HTML or XML document, expands doclets,
    and returns final string.  Ideally, it should affect nothing
    but the doclets.  In practice this depends on the python parser,
    and we'll settle for 'no serious change'.

    """

    def asDict(self, attrs):
        d = {}
        for (k,v) in attrs:
            d[k] = v
        return d

    def process(self, source, docletMap, nameSpace, tempDir):
        "Do it all for me"
        self.reset()

        self._docletMap = docletMap
        self._nameSpace = nameSpace
        self._output = []
        
        self.feed(source)
        self.close()

        return self.getOutput()
    

    def getOutput(self):
        return ''.join(self._output)

    def handle_starttag(self, tag, attrs):
        #is it a special doclet?
        if tag in self._docletMap:
            self.handle_start_doclet(tag, attrs)
        else:
            out = ['<%s' % tag]
            for (k, v) in attrs:
                out.append('%s="%s"' % (k, v))
            out.append('>')
            self._output.append(' '.join(out))

    def handle_endtag(self, tag):
        if tag in self._docletMap:
            self.handle_end_doclet(tag)
        else:
            self._output.append("</%s>" % tag)

    def handle_data(self, data):
        #assume the data was just parsed, cdata not handled yet.
        self._output.append(data)

    def handle_charref(self, name):
        self._output.append(name)

    def handle_entityref(self, name):
        self._output.append(name)

    def handle_comment(self, data):
        self._output.append('<!--%s-->' % data)
        
    def handle_decl(self, decl):
        self._output.append('<!%s>' % decl)

    def handle_pi(self, data):
        self._output.append('<?%s>' % data)


    def handle_start_doclet(self, tag, attrs):
        #attributes override namespace and go into
        # dict passed to setData
        print('doclet %s, attrs=%s' % (tag, str(attrs)))
        if attrs:
            ns = self._nameSpace.copy()
            for (key, value) in attrs:
                ns[key] = value
        else:
            ns = self._nameSpace

        #create the thing
        klass = self._docletMap[tag]
        objekt = klass()

        #let it pull data
        objekt.setData(tag, self.asDict(attrs), ns)

        #make the HTML
        html = objekt.asHTML()

        self._output.append(html)

    def handle_end_doclet(self, tag):
        pass
    


def testRml():
    from reportlab.platypus.doctemplate import SimpleDocTemplate
    from reportlab.platypus.flowables import Spacer
    from reportlab.lib.randomtext import randomText

    templ = SimpleDocTemplate('doclet_output.pdf')

    #need a style
    story = []
    normal = ParagraphStyle('normal')
    normal.firstLineIndent = 18
    normal.spaceBefore = 6

    para = Paragraph("Test of doclets.  You should see a little table with a reversed word.", normal)
    story.append(para)

    story.append(Spacer(36,36))
                 
    theDoclet = TestReverseDoclet()
    story.append(theDoclet.asFlowable())
    
    for i in range(5):
        para = Paragraph(randomText(), normal)
        story.append(para)



    templ.build(story)    

    print('saved doclet_output.pdf')

    
if __name__=='__main__':
    #make a platypus doc and an HTML one with the
    #test objects
    testRml()



#copyright ReportLab Inc. 2000-2016
"""ReportLab Application Monitor.

This provides utilities to help manage and verify software installations.
The basic problems are that source and binary distriubutions can get
messed up at many points in between our CVS repository and a customer's
live site.  So we need a system to verify integrity.  This should be
easy for customers' configuration management staff to understand.

We have a class which sits over a directory and knows how to do
various things to the files under it.  This can be subclassed
whenever the desired behaviour is not quite right.  The base
class should be good enough to do the right things for standard
ReportLab apps with python and preppy files.  Main functions
are
- help put together a manifest file in development
- generate a signed, checkable 'source manifest' when we
  ship source
- verify the above in a custromer's source environment
  and/or warn of changes
- compile the app, generating pyc files etc.
- generate a signed and checkable 'compiled manifest'
  at the moment of compilation
- if running in live, verify the signatures of the
  compiled files
- warn of stuff which perhaps should not be there,
  e.g. source files in a compiled environment, extra
  modules which should be deprecated etc.

An 'application' will have a manifest we maintain in version
control which lists the files that matter.  By convention
this is filelist.txt.

When we ship something, we can include a manifest_source.txt
with checksums.  This is NOT version controlled - it cannnot
be as checking it in will change its checksum :-)  It can
be generated by a script which makes the zip/tgz for
the customer.

When they run the compile action, it generates a manifest_compiled.txt
with checksums for all the pycs and binary files.  This is promoted
live with the rest of the code.
"""
from __future__ import print_function
__version__='3.3.0'
import sys, os, time, string

import preppy
import pprint
import py_compile
import stat
import shutil
import zipfile
from rlextra.utils import gloop
from rlextra.utils.manifester import Manifester

def getCVSFilesInDirectory(directory):
    """Helper for a method defined below.  Returns a list of the files
    which are listed in the CVS entries.  Barfs if not all are present."""
    if not os.path.isdir(directory + os.sep + 'CVS'):
        return []

    cvsEntries = open(directory + os.sep + 'CVS' + os.sep + 'Entries').readlines()
    filenames = []
    for entry in cvsEntries:
        if len(entry) > 0:
            words = string.split(entry, '/')
            if len(words) > 1:
                filename = words[1]
                fullName = directory + os.sep + filename
                assert os.path.exists(fullName), 'file "%s" listed in CVS but not present!' % fullName
                if os.path.isfile(fullName):
                    filenames.append(fullName)
    filenames.sort()
    return filenames

class AppMonitor(Manifester):
    flName  = 'filelist.txt'                        #name of the original packing list
    ssmName = 'source_manifest.txt'                 #name of the signed source manifest
    stmName = 'target_manifest.txt'                 #name of the signed target manifest
    _force_ssmName = 1                              #force ssmName into target list
    # these may have line ending variations
    def __init__(self, dirname=None, id='', verbose=0, fileList=None):
        Manifester.__init__(self,dirname=dirname,id=id,verbose=verbose)

        self._sourceFiles = fileList
        # when checks fail, track the list of filenames here
        # but continue to a full report.
        self.source_failed = []
        self.source_missing = []
        self.target_failed = []
        self.target_missing = []
        self.unexpected = []

        #can optionally supply a list of filenames, instead of
        #reading a list file name at top level.
        self.suppliedFileList = fileList

    def getList(self):
        "Returns list of files"
        if not self._sourceFiles:
            #read filelist.txt at top level
            lines = open(self.flName,"r").readlines()
            self._sourceFiles = self.stripComments(lines)
        return self._sourceFiles

    def writeSignedSourceManifest(self):
        "Generates signed source manifest"
        self.writeSignedManifest(self.getList(),self.ssmName,_id=1)

    def verifySignedSourceManifest(self):
        "Reads and checks against signed source manifest"
        result, self.source_failed, self.source_missing = self.verifySignedManifest(self.ssmName)
        return result

    def compilePy(self, filename, force=0, verbose=0):
        "Recompile a python module if necessary."
        if verbose: print('compilePy', filename, '...', end=' ')
        cfile = filename + (__debug__ and 'c' or 'o')
        ftime = os.stat(filename)[stat.ST_MTIME]
        try: ctime = os.stat(cfile)[stat.ST_MTIME]
        except os.error: ctime = 0
        if (ctime > ftime) and not force:
            if verbose: print('up-to-date')
            return
        else:
            py_compile.compile(filename)
            if verbose: print('recompiled',cfile)


    def compilePrep(self, filename, force=0, verbose=0):
        "Recompile a preppy module if necessary."
        # todo - harmonize print output and force with Python
        # modules
        preppy.compileModule(filename)

    def compileApp(self, force=0):
        "Default compilation for prep and py files in source list"
        for filename in self.getList():
            root, ext = os.path.splitext(filename)
            # cute - if ends in .py, look for compilePy;
            # if ends in spam, look for compileSpam. You
            # can subclass for new file types
            try:
                compileMethodName = 'compile' + ext[1].upper() + ext[2:]
                compileMethod = getattr(self, compileMethodName, None)
                if compileMethod:
                    compileMethod(filename, force=force)
            except:
                print("Can't compile '%s'" % filename)

    def getCompiledFileName(self, filename):
        "What depends on this? Used to build target file list"
        # default is a 1-to-1 correspondence but we should
        # allow for one source having multiple outputs.
        # this returns either a single name or a list of names.
        root, ext = os.path.splitext(filename)
        if ext == '.py':
            compiled = filename + (__debug__ and 'c' or 'o')
        elif ext == '.prep':
            # I guess pyo files are a possibility...
            compiled = root + '.py' + (__debug__ and 'c' or 'o')
        else:
            compiled = filename
        return compiled

    def getTargetFileList(self):
        "Work out the list of filenames expected in the target installation"
        sourceFiles = self.getList()
        # we may get dependencies - one-to-many or many-to-one
        # so build a set of outputs and sort them at the end.

        targets = {}
        for sourceFile in sourceFiles:
            comp = self.getCompiledFileName(sourceFile)
            # it might return a string, or a list of more than one name
            if type(comp) == type(''):
                targets[comp] = 1
            else:  # got a list of stuff
                for outfile in comp:
                    targets[outfile] = 1

        # the default sort would mix files and directories;
        # a.text, b/subdir.txt, c.txt.  So sort in a 'pretty'
        # manner - files and directories.
        tmp = []
        T = list(targets.keys())
        if self._force_ssmName and self.ssmName not in T: T.append(self.ssmName)    #force add source manifest
        for filename in T:
            directory, shortname = os.path.split(filename)
            tmp.append((directory, shortname, filename))
        tmp.sort()
        outfiles = []
        for (a,b,c) in tmp:
            outfiles.append(c)

        return outfiles

    def writeSignedTargetManifest(self):
        "Generates signed target manifest for compiled application"
        self.writeSignedManifest(self.getTargetFileList(),self.stmName)

    def verifySignedTargetManifest(self):
        "Reads and checks against signed target manifest"
        result, self.target_failed, self.target_missing = self.verifySignedManifest(self.stmName)
        return result

    def findSourceFilesInTarget(self):
        "Hunt for uncompiled files e.g on a public web server"
        # intent is to find .py and .prep files.  Be cleverer -
        # - there will be complex dependency issues e.g. with
        # pagecatcher forms. The rule: if it is present in sourcefiles
        # and absent in target file list, assume it should not be there.
        strays = []

        sourceFiles = self.getList()
        targetFiles = self.getTargetFileList()

        targetTest = {} # hash for quick testing of presence
        for targetFile in targetFiles:
            targetTest[targetFile] = 1

        for srcFile in sourceFiles:
            if srcFile in targetTest:
                pass # OK, same in both versions e.g. HTML file
            else:
                if os.path.isfile(srcFile):
                    strays.append(srcFile)
        return strays

    def findUnexpectedFilesInTarget(self, exceptions=[]):
        """Finds any files apparently not from the app at all.
        Checks any directory for which there are any files listed.
        Subdirectories (logs, output etc.) are not scanned. Also
        you may give a list of known exceptions which are OK,
        for things the app generates locally."""
        unexpected = []
        doneDirs = {}
        sourceFiles = self.getList()
        targetFiles = self.getTargetFileList()

        # make some dictionaries for fast testing
        sourceSet = {}
        for src in sourceFiles:
            sourceSet[src] = 1
        targetSet = {}
        for targ in targetFiles:
            targetSet[targ] = 1

        for targetFileName in self.getTargetFileList():
            dirname, filename = os.path.split(targetFileName)
            if dirname in doneDirs:
                # we did it already, skip
                continue
            else:
                # scan for anything unknown in this directory
                doneDirs[dirname] = 1
                if dirname == '':
                    filelist = os.listdir('.')
                else:
                    filelist = os.listdir(dirname)

                for filename in filelist:
                    # windows will use 'parent\\subdir', but we want to
                    # test if it is in our list as 'parent/subdir'
                    normalized = string.replace(os.path.join(dirname, filename),
                                                os.sep, '/')
                    if filename in [self.ssmName, self.stmName]:
                        pass
                    elif normalized in sourceSet:
                        pass
                    elif normalized in targetSet:
                        pass
                    elif os.path.isdir(normalized):
                        pass
                    else:
                        unexpected.append(normalized)

        return unexpected

    def promoteTargetFiles(self, destDir, createDest=0):
        "Copies them up to destDir if changed"
        if not os.path.isdir(destDir):
            if createDest:
                os.makedirs(destDir)
            else:
                raise IOError("destination directory '%s'not found, please check spelling!" % destDir)

        for targetName in self.getTargetFileList():
            dirname, filename = os.path.split(targetName)
            if not os.path.isdir(os.path.join(destDir, dirname)):
                os.makedirs(os.path.join(destDir, dirname))
            shutil.copy2(os.path.join(dirname,filename), os.path.join(destDir, dirname))
            print('copied',os.path.join(destDir, dirname, filename))

        # and finally the manifest
        shutil.copy2(self.stmName, destDir)

    def promoteSourceFiles(self, destDir, createDest=0):
        "Copies them up to destDir, preserving timestamps"
        if not os.path.isdir(destDir):
            if createDest:
                os.makedirs(destDir)
            else:
                raise IOError("destination directory '%s'not found, please check spelling!" % destDir)

        for targetName in self.getList():
            dirname, filename = os.path.split(targetName)
            if not os.path.isdir(os.path.join(destDir, dirname)):
                os.makedirs(os.path.join(destDir, dirname))
            shutil.copy2(os.path.join(dirname,filename), os.path.join(destDir, dirname))
            print('copied',os.path.join(destDir, dirname, filename))

        # and finally the manifest
        shutil.copy2(self.ssmName, destDir)

    def promoteViaFTP(self):
        "Moves compiled code onto another box..."
        raise NotImplementedError

    def _writeArchive(self,arch,filelist,copyDir=None):
        if copyDir:
            from .buildutils import copyTree, kill
            if os.path.isdir(copyDir): kill(copyDir)
            if self.verbose: print('Copying to directory "%s"' % copyDir)
            copyTree('.',copyDir,excludeSVN=1,exclude=[copyDir])
            filelist = list(map(lambda x, copyDir=copyDir: os.path.join(copyDir,x), filelist))
        for filename in filelist:
            arch.write(filename)
            if self.verbose: print('archiving',filename)
        if copyDir: kill(copyDir)
        arch.close()

    def makeArchive(self, archiveName, filelist, kind='zip', prefix=None, copyDir=None):
        """Create a zip or tgz containing given files.  Uses tarfile
        module from 3rd party which will hopefully go into Python 2.3 one
        day; thanks to author!"""
        if kind is None:
            # infer from extension.  .tar.gz is not handled by splitext;
            # check ourselves
            if archiveName[-4:] == '.zip':
                kind = 'zip'
            elif archiveName[-4:] == '.tar':
                kind = 'tar'
            elif archiveName[-4:] == '.tgz':
                kind = 'tgz'
            elif archiveName[-7:] == '.tar.gz':
                kind = 'tar.gz'
            elif archiveName[-7:] == '.pyz':
                kind = 'pyz'

        if kind in ['tgz', 'tar.gz']:
            # the tar module I found is Python 2.2 only and chokes
            # when gzipping; had to handle that myself
            from rlextra.utils import tarfile
            import tempfile
            tarName = tempfile.mktemp()
            arch = tarfile.TarFileCompat(tarName, 'w', compression=tarfile.TAR_PLAIN)
            self._writeArchive(arch,filelist,copyDir=copyDir)

            import gzip
            z = open(archiveName,'wb')
            g = gzip.GzipFile(archiveName[:-(len(kind)+1)]+'.tar', 'wb', 9, z)
            f = open(tarName, 'rb')
            while 1:
                chunk = f.read(1024)
                if not chunk:
                    break
                g.write(chunk)
            g.close()
            z.close()
            f.close()
            os.remove(tarName)
        elif kind =='tar':
            from rlextra.utils import tarfile
            arch = tarfile.TarFileCompat(archiveName, 'w', compression=tarfile.TAR_PLAIN)
            self._writeArchive(arch,filelist,copyDir=copyDir)
        elif kind == 'zip':
            arch = zipfile.ZipFile(archiveName, 'w', compression=zipfile.ZIP_DEFLATED)
            self._writeArchive(arch,filelist,copyDir=copyDir)
        elif kind == 'pyz':
            # ensure installer on path
            import rlextra
            instDir = os.path.abspath(os.path.join(os.path.dirname(rlextra.__file__),'distro','tools','Installer'))
            print('instDir = %s' % instDir)
            if instDir not in sys.path:
                sys.path.insert(0,instDir)
            print('added directory %s to path' % instDir)
            from MEInc.Dist.archive import ZlibArchive
            from py_compile import compile
            instDir = os.path.join(instDir, 'support')
            # Z is the table of contents for the archive
            Z = []
            #by default add the unarchive code
            preLoad=1
            if preLoad:
                for n in ('archive_rt','imputil'):
                    f = os.path.join(instDir,n+'.py')
                    c = f+'c'
                    compile(f,c,f)
                    Z.append((n,c))
            for fileName in filelist:
                #work out the 'import name'
                nameMinusExt = os.path.splitext(fileName)[0]
                importName = string.replace(nameMinusExt, os.sep, '.')
                importName = string.replace(importName, '/', '.')
                print('store %s as %s' % (fileName, importName))
                Z.append((importName, fileName))
            arch = ZlibArchive().build(archiveName, Z)
            #def makePYZ(where, pyz_exclude=(), remove_after_pyz=(), pyzName=None,
            # preLoad=1, addLib=1, cleanUp=1):

        else:
            raise ValueError("unknown format, use 'zip' or 'tgz' or 'pyz' only")

        if self.verbose: print('wrote',archiveName)
        return os.stat(archiveName)[stat.ST_SIZE]

    def makeSourceArchive(self, archiveName, kind=None, copyDir=None):
        "Create a zip or tgz containing all source files"
        files = self.getList()
        files.append(self.ssmName)
        return self.makeArchive(archiveName, files, kind=kind, copyDir=copyDir)

    def makeTargetArchive(self, archiveName, kind=None):
        "Create a zip or tgz containing all compiled files"
        files = self.getTargetFileList()
        files.append(self.stmName)
        return self.makeArchive(archiveName, files, kind=kind)

    def targetHtmlReport(self, encrypted=1):
        "Verify a target installation"
        result = self.verifySignedTargetManifest()
        sources = self.findSourceFilesInTarget()
        unexp = self.findUnexpectedFilesInTarget()

        if result:
            print('status:  ok')
        else:
            print('status: bad', time.time())

        stuff = []
        L = stuff.append
        if len(self.target_failed):
            L('%d corrupted files:' % len(self.target_failed))
            for filename in self.target_failed:
                L('    ' + filename)
            L('')

        if len(self.target_missing):
            L('%d missing files:' % len(self.target_missing))
            for filename in self.target_missing:
                L('    ' + filename)
            L('')

        if len(sources):
            L('%d unexpected source files:' % len(sources))
            for filename in sources:
                L('    ' + filename)
            L('')

        if len(unexp):
            L('%d totally unexpected files:' % len(unexp))
            for filename in unexp:
                L('    ' + filename)
            L('')

        if stuff or not result:
            M = self._getSignedManifest(self.ssmName)
            if M:
                L('\n\nSource Manifest\n===============')
                list(map(L,M))

        if encrypted:
            print(gloop.Gloop('hex').encode('\n'.join(stuff)))
            print()
        else:
            for line in stuff:
                print(line)

    def getCVSFilesUnderMe(self):
        """Returns a list of all CVS-committed files below directory"""
        filelist = []
        def walker(filelist, dirname, names):
            filelist.extend(getCVSFilesInDirectory(dirname))
        os.path.walk(self.baseDir, walker, filelist)
        # if they begin with .\ or ./, trim it
        cleaned = []
        for filename in filelist:
            if filename[0:2] == os.curdir + os.sep:
                filename = filename[2:]
            cleaned.append(filename)
        return cleaned

def test(id,D,verbose=1,ws=1,cs=1):
    import pprint
    #need to point this at a non-customer demo later
    a = AppMonitor(id,D,verbose)
    if ws:
        a.writeSignedSourceManifest()
        print()
    if cs:
        print('checking source_manifest...')
        a.verifySignedSourceManifest()
        if a.source_failed and verbose:
            print('Failed')
            if verbose>1:
                pprint.pprint(a.source_failed)
    if verbose: print('done')

if __name__=='__main__':
    # robin, apologies, dunno if mine is better but it helps
    # test - you can call any method name from the command line
    # directly and just point it at a directory.

    args = sys.argv[:]

    if '-v' in args:
        args.remove('-v')
        verbose = 1
    else:
        verbose = 0

    if len(args) < 2:
        print('Usage:  appmonitor.py directory methodname [arg1] [arg2] [arg3]...')
        print('n.b.    appmonitor.py ../myapp targetHtmlReport produces HTML')
    else:
        directory = args[1]
        if not os.path.isdir(directory):
            print('directory %s not found, try again!' % directory)
            sys.exit()

        methodName = args[2]
        if methodName == 'targetHtmlReport':
            print('Content-type: text/html')
            print()
            print('<html><head><title>%s</title></head><body><pre>' % methodName)

        args = args[3:]
        a = AppMonitor(dirname=directory, verbose=verbose)
        methObj = getattr(a, methodName, None)
        if methObj is None:
            print('command "%s" not found!' % methodName)
        else:
            retVal = methObj(*args)
            if retVal is None:
                print('Completed.')
            else:
                print(' Completed.  Return value:')
                pprint.pprint(retVal)
                if methodName[0:6] == 'verify' and retVal == 0:
                    print('source_failed = %s' % a.source_failed)
                    print('source_missing = %s' % a.source_missing)
                    print('target_failed = %s' % a.target_failed)
                    print('target_missing = %s' % a.target_missing)


        if methodName == 'targetHtmlReport':
            print('</pre></body></html>')



"""
pageCatcher demo script:
   This version of "append" uses a specialized copy feature of pageCatcher
   which does not permit the modification of pages.  However, all pages of
   the copied document will be put into the output document together with
   annotations, bookmarks and outlines.
"""

try:
    from rlextra.pageCatcher.pageCatcher import copyPages
except ImportError:
    pass # running inside pageCatcher module?

from reportlab.pdfgen import canvas

def doappend(topdffile, frompdffilelist):
    canv = canvas.Canvas(topdffile)
    for frompdffile in frompdffilelist:
        print("copying", frompdffile)
        copyPages(frompdffile, canv)
    print("\n\nnow writing", topdffile)
    canv.save()

if __name__=="__main__":
    # edit this
    doappend("out8_directcopy.pdf", ["sample1.pdf", "sample2.pdf", "sample3.pdf"])


"""
Basic tests of embedding & restoring forms
"""


import string, os, sys

import unittest
from reportlab.lib.testutils import ScriptThatMakesFileTest
from rlextra.pageCatcher import pageCatcher
from reportlab.pdfgen.canvas import Canvas

global VERBOSE
VERBOSE = 0

# locations must be relative to reportlab package for this test to work, or absolute

class PageCatcherTests(unittest.TestCase):
    """Various tests of pageCatcher's capabilities"""

    def setUp(self):
        #check we're in the pagecatcher directory
        self.startDir = os.getcwd()
        cwd = os.getcwd()
        
        if not cwd.endswith('pagecatcher'):
            if os.path.isdir('pagecatcher'):
                os.chdir('pagecatcher')
    
    
        if not os.path.isfile('sample.frm'):
            pageCatcher.storeForms('sample1.pdf',
                                   'sample1.frm',
                                   prefix='sample1_page',
                                   all=1)
            if VERBOSE: print('cached sample1.frm for later tests')

    def tearDown(self):
        os.chdir(self.startDir)

    def testRestore(self):
        "store and restore a form"
        c = Canvas('test_pc_1.pdf')
        c.setFont('Helvetica-Bold',24)
        c.drawString(100,700, 'Testing how forms are restored and used')
        names = pageCatcher.restoreForms('sample1.frm',c)
        if VERBOSE: print('restored %s' % names)
        c.scale(0.5, 0.5)
        c.translate(200,600)
        c.doForm('sample1_page0')
        #dumpObjects(c)
        if VERBOSE: print('drew page 1 of sample')
        c.translate(0,-400)
        c.doForm('sample1_page1')
        if VERBOSE: print('drew page 2 of sample')
        c.save()
        if VERBOSE: print('saved test_pc_1.pdf')

    def xxtestDrawBeforeRestore(self):
        "draw forms before restoring them form"
        c = Canvas('test_pc_2.pdf')
        c.setFont('Helvetica-Bold',24)
        c.drawString(100,700, 'Testing draw before restore')

        c.scale(0.5, 0.5)
        c.translate(200,600)
        c.doForm('sample1_page0')
        if VERBOSE: print('drew page 1 of sample')
        c.translate(0,-400)
        c.doForm('sample1_page1')
        if VERBOSE: print('drew page 2 of sample')
        names = pageCatcher.restoreForms('sample1.frm',c)
        if VERBOSE: print('restored %s' % names)
        c.save()
        if VERBOSE: print('saved test_pc_2.pdf')


    def testDrawNonExistent(self):
        "Draw a form we have not restored. Should raise error"
        c = Canvas('test_pc_3.pdf')
        c.setFont('Helvetica-Bold',24)
        c.drawString(100,700, 'Testing draw before restore')
        c.scale(0.5, 0.5)
        c.translate(200,600)
        c.doForm('sample1_page0')
        if VERBOSE: print('drew page 1 of sample')
        self.failUnlessRaises(KeyError, c.save)
        if VERBOSE: print('saved test_pc_3.pdf, oh dear')

    def testDoubleRestore(self):
        "Restore Twice. Should raise error"
        c = Canvas('test_pc_4.pdf')
        c.setFont('Helvetica-Bold',24)
        c.drawString(100,700, 'Testing double restore')
        c.scale(0.5, 0.5)
        c.translate(200,600)
        names = pageCatcher.restoreForms('sample1.frm',c)
        if VERBOSE: print('double restore coming...')
        self.failUnlessRaises(ValueError, pageCatcher.restoreForms, 'sample1.frm',c)

    def testCombinePdfsInMemory(self):
        "Merge 3 files using combinePdfInMemory"
        if os.path.isfile('combined_mem.pdf'):
            os.remove('combined_mem.pdf')

        stuff = []
        stuff.append(open('sample1.pdf','rb').read())
        stuff.append(open('sample2.pdf','rb').read())
        stuff.append(open('sample3.pdf','rb').read())
        pdf = pageCatcher.combinePdfsInMemory(stuff)

        open('combined_mem.pdf','wb').write(pdf)
        if VERBOSE: print('saved combined_mem.pdf')

    def testCombinePdfsOnDisk(self):
        "Merge 3 files using combinePdf"
        if os.path.isfile('combined_disk.pdf'):
            os.remove('combined_disk.pdf')
        stuff = []
        stuff.append('sample1.pdf')
        stuff.append('sample2.pdf')
        stuff.append('sample3.pdf')
        pageCatcher.combinePdfs('combined_disk.pdf',stuff)
        assert os.path.isfile('combined_disk.pdf')
        if VERBOSE: print('saved combined_disk.pdf')

    def testNeedsPyPdf(self):
        with open('needs-pypdf.pdf','rb') as f:
            pdfContent = f.read()
        pageCatcher.storeFormsInMemory(pdfContent,all=True)
        if VERBOSE: print('Stored needs-pypdf.pdf in memory')

def dumpObjects(canvas):
    """Show the PDF objects within"""
    doc = canvas._doc
    import pprint
    print('forms:')
    for (key, value) in list(doc.idToObject.items()):
        if key[0:7] == 'FormXob':
            print('   ',key,'=',value)

def makeSuite():
    return unittest.TestSuite((
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example1_fillform.py',
##                               'out1_fillform.pdf',
##                                verbose=VERBOSE),
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example2_reverse.py',
##                               'out2_reversed.pdf',
##                                verbose=VERBOSE),
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example3_append.py',
##                               'out3_combined.pdf',
##                                verbose=VERBOSE),
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example4_fourpage.py',
##                               'out4_fourpage.pdf',
##                                verbose=VERBOSE),
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example5_background.py',
##                               'out5_background.pdf',
##                                verbose=VERBOSE),
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example6_twopage.py',
##                               'out6_twopage.pdf',
##                                verbose=VERBOSE),
##                ScriptThatMakesFileTest('../rlextra/pageCatcher/demos',
##                               'example7_split.py',
##                               'out7_page9.pdf',
##                                verbose=VERBOSE),
                unittest.makeSuite(PageCatcherTests,'test'),
                )
        )

#noruntests
if __name__ == "__main__":
    VERBOSE = ('-v' in sys.argv)
    w = os.path.dirname(sys.argv[0])
    if w: os.chdir(w)
    unittest.TextTestRunner().run(makeSuite())



#Autogenerated by ReportLab guiedit do not edit
from reportlab.graphics.widgets.grids import ShadedRect
from reportlab.graphics.shapes import Drawing, _DrawingEditorMixin, String
from reportlab.lib.colors import grey

class Graphic(_DrawingEditorMixin,Drawing):
    def __init__(self,width=400,height=200,*args,**kw):
        Drawing.__init__(self,width,height,*args,**kw)
        self.width       = 72
        self.height      = 20
        self._add(self,ShadedRect(),name='sr',validate=None,desc=None)
        self.sr.strokeColor    = None
        self.sr.fillColorEnd   = grey
        self.sr.width=self.width
        self.sr.height=self.height
        self._add(self,String(self.width*0.5,6,'text', textAnchor='middle', fontName='Helvetica-Bold',fontSize=12),name='s',validate=None,desc=None)

if __name__=="__main__": #NORUNTESTS
    Graphic().save(formats=['pdf'],outDir='.',fnRoot=None)



__code__='Gas1]9lJKG&Daab%n^P/#ebd?TEle$5RR`=KRbm%Ua/nK:h2hWj5%/@NbrquD/+c\'HKBqOQ[In%E+6g]g<]n@p1_/qo4DVmcTfP$miniGY[e$T/j&,/@?cmA_T58h.&S<Ld(p*m0obGA?j0SA-!"X@fF0jfB%[>Ga6)%V]D-gRM`r>#jE(!XjqpjE(HLR]08t!F^92=9O2f3AX0>6P`$qs+LKKHPh=RV+U/-fT#f*,b,E;sSU$s4JF@IFQ+=K5(^-9gs`iG,FX4RIodRE_$=(FjV3n$R0>ZN\\jFZbQ0L0A\'L?96M(PX;.\\CUDR^mZM\\D2dWuN2V)3K[IgQ-c\'>2ka@^Jh\'G;!Gatb_>no-+S;5X*K[\\Iu?+%2?eqCbKpI%Z.:B>qT\'Y<8mXiRc(R[Cuu7=mL:lOP.`AVBlOl+Yf&)17a+H2)ahIZ"P(kP[-6=HJEYL);"8pX<40"lagI#??kNC1Q\\%ASq;"hGY8s\\,Euor1,jFj$e=m\\k\'HW)`sed.R:5%F89/;h!r.1T``lA]4%^f[CUO->I9Yd%U"?^N\\7_;1h:V>9^8HQB11Ob`5nXHrn`/]g:CaH0n[Ieg&H,!k.fMNXS4*?4iqfWLp:-3/%Md$]i\'=Igkh=Mea%U-%QKrOd4QuX[F6cN+(@[ba?hc^SjL-kRI.uK\'poiP$E.<#l;-H!%I2sah+tlW.#!La\'TuD?&ET<WIa(Dp@mr3b8Lpe\\?P#`p]1V(jYR9#BP%&;\'!0*jR2B-t7\\DHc:D^FoY2@(m020^m\\`GP(\'h%R@Yj(][\'.c`r2VC8,biT[cMV#Fg\\`FYN.FBf=n_=au:j)ru<p/6*[li^@fG!!ZL>q*IsShb[WIjl-A^BqYCpk)mGuV":S(\\RpDr@!Y\\[!XI5L)\',4ciN28*C:f(/N4eX2\'apshMrm]63@hYC(PY2?1E4E)\'`u@$18DH!h%Q`k4R$c6ai[D>FS-?sj_ap]d(*VJ)H?U[;q2Y^U&>0OdTiL~>'



__code__='Gat=(s)6@U\'RuIZTe-[Fl.FWrG^S2]UcC[XS#0m<$gEE,Aecb*$C`Tun(O3k[di1]WDn<rdXKkV[D-PHSSaN"Bh3+$GP9k+c^/;%MdY]93aFttV>d0=2dljY$OT.L2o+sn!@%VC^bP?\'j=cT(5l*>=*Pa\\N(>p(GH,\'U!7DQ6[k5[^H4:cZ;E!c"Xi%j4:^_OWq!A=P("Sa_0\'u3TF56\\\\,GRa-u_!ApZiFS4`*rgL#rZH(`7WT\\VX?\\fhkMWiVimlp?I,VjK@Wqi;B;m-(E!$=3>@bVW$H#D6(3s`EoeY@88`9rJ-`nCE3%DTI2YBNn+[$=hc4m0ZiUG>:?!YJ?\\Qf2r"C-E3"DpGcRLDLaeZYVboh<h,ed`ZW+Fd@/\'jTF%!@"p_/d_F;/B-PU$9"J>l#68:YAL_9d@?o7,B]r-@<c659U)ojWg`&Ac,*#`S/61F9+-d9PAFGVWF7Gi[::+3DP_3YAV7N\\Th+64(ZEuHh?.]hgQ!bX:,UQV3_iYpCh3p!q4UO/Es>S7[`")5=Y\'m&2Y?RQ?cqN=+@sDT$j*bK+3#1eKHJOna&aT]Wc_Nr/<prpRfukb8J[HI+?6c8q]ZnZE?Lr+rR#"\\enK5uauh2]+)\\V.rj..p.lVR630L8(kiN$tNjP9mcg9E:3rcqC3B0"f$9OZ/>@*PR4h!iIqo5D9[iZqDe8^ec=lILp*Z\'JHAbk:V32WNOVbHam>O5rj@<,W]H#Ha8\\2.:VgV%M"+T0:T^+9t6eQM#MP%aOCKQZ2-3,,=#G__sa^L?P725q(bG101QG=m%6khOr-d;GnY=/N1qc$:5rhn=6#ILkmqoAJTmGLb+tZ[>+8P>o^L-!=8^n+X!\\Jo9R&Y>:!XS6o8O<,Qh_6?[d`6t<SSUV/YCW@M:-qE<#\'^P2M)hGtqphd&#g1[>gph%&I5@Kl7jrJ9bKfj=@=4h%p\')4;]"L[Oan95J%3\\hiq:gT\'!;Inc+[\\Tth%[PtpTWl=25qf"c6:iBTlTku3Jhk0.CeS2r]dWF4"[Y`\\o;P#lYP>`P\\Q5j#fCFn8G<p7IrkPJ2H@=,_+Da=p]#mi,iEZ-I93b\'b^atE&6\\::7E9nB42:EuBpr(si"+.5QG`@jJ]C+Y5FdJBL8.ub:kCA\\H=$_k0~>'



from reportlab.graphics.charts.textlabels import Label
from reportlab.graphics.shapes import Drawing, Group, Line
from reportlab.graphics.charts.legends import Legend
from reportlab.graphics.widgetbase import TypedPropertyCollection
from reportlab.lib.colors import PCMYKColor, _PCMYK_black
from reportlab.lib.attrmap import *
from reportlab.lib.units import cm
from reportlab.lib.validators import *
from reportlab.pdfbase.pdfmetrics import getFont
from reportlab.graphics.charts import barcharts
from reportlab.graphics.charts.lineplots import _maxWidth

def _highest(data, roundInterval=10):
    v=max(list(map(max,data)))
    d,f=divmod(v,roundInterval)
    if f>0:
        f=1
    else:
        f=0.5
    return (d+f)*roundInterval

class VerticalBarChart(barcharts.VerticalBarChart):
    """Returns a widget - used by the class VerticalBarChartDrawing.
    Based on (and inherits from) barcharts.VerticalBarChart. Each bar has a vertical
    line from the (x)centre to the its label. The properties to do with
    this line are self.lineLength, self.lineStrokeWidth and self.lineColor.
    All the labels are at the same height, (ie in a horizontal line _outside_
    their bars).
    It's used for Chartbook 2001 - plot for eg Performance Year by Year."""

    _attrMap = AttrMap(BASE=barcharts.VerticalBarChart,
        labelOffset = AttrMapValue(isNumber,
            desc="Space between label text and grid edge"),

        background = AttrMapValue(isColorOrNone,
            desc="Color for the plot area background (if any)"),
        categoryNames = AttrMapValue(isListOfStrings,
            desc="Group names (eg years)"),

        lineLength = AttrMapValue(isNumber,
            desc='Length of the horizontal line (behind the actual bars)'),
        lineStrokeWidth = AttrMapValue(None,
            desc='Stroke width of the horizontal line (behind the actual bars)'),
        lineColor = AttrMapValue(isColorOrNone,
            desc='Colour of the horizontal line (behind the actual bars)'),
                       
        barLabelFormat = AttrMapValue(isAnything,
            desc="Formatting routine to be used on the bar labels"),
        barFillColors = AttrMapValue(isListOfColors,
            desc="List of colours to be used filling in the bars"),
        legend = AttrMapValue(None,
            desc="Handle to the chart legend"),

        leftPadding = AttrMapValue(isNumber,
            desc='Padding on left of drawing'),
        rightPadding = AttrMapValue(isNumber,
            desc='Padding on right of drawing'),
        topPadding = AttrMapValue(isNumber,
            desc='Padding at top of drawing'),
        bottomPadding = AttrMapValue(isNumber,
            desc='Padding at bottom of drawing'),
        )
    
    def __init__(self):
        barcharts.VerticalBarChart.__init__(self)
        self.height = 2.8*cm # height of BARS/LINES - NOT the whole chart
        self.width = 4.8*cm # width of BARS/LINES - NOT the whole chart
        self.background = None
        self.data      = [(24.7, 18.2, 23.9, 16.1, 13.8),
                          (23.8, 15.6, 23.8, 14.9, 25.1)]
        self.barLabelFormat = '%.1f%%'

        # used for the vertical line behind the bars
        self.lineLength = 2.8*cm
        self.lineStrokeWidth = 0.5
        self.lineColor = _PCMYK_black
 
        self.bars.strokeColor = _PCMYK_black
        self.bars.strokeWidth = 0.5
        self.groupSpacing = 0.9*cm
        self.barSpacing = 0.06*cm
        self.barWidth = 0.19*cm
        
        #Allows us up to six bars (per group) to be coloured
        c0 = PCMYKColor(100.0,65.0,0.0,30.0,density=100)
        c1 = PCMYKColor(11.0,11.0,72.0,0.0,density=100)
        self.bars[0].fillColor=c0
        self.bars[1].fillColor=c1
        self.bars[2].fillColor=PCMYKColor(100.0,65.0,0.0,30.0,density=75)
        self.bars[3].fillColor=PCMYKColor(11.0,11.0,72.0,0.0,density=75)
        self.bars[4].fillColor=PCMYKColor(100.0,65.0,0.0,30.0,density=50)
        self.bars[5].fillColor=PCMYKColor(11.0,11.0,72.0,0.0,density=50)

        self.valueAxis.valueMin = 0.0
        tl=[]
        for f in range (0, len(self.data)):
            for i in range (0,len(self.data[f])):
                tl.append(self.data[f][i])
        self.valueAxis.valueMax = _highest([tl])
        
        self.categoryAxis.visibleAxis = 0
        self.categoryAxis.visibleTicks = 1
        self.categoryAxis.tickDown = 0

        self.barLabels.fontName = "Helvetica-Oblique"
        self.barLabels.fontSize = 6
        self.barLabels.fillColor = _PCMYK_black
        self.barLabels.angle = 90

        self.labelOffset = 5

        self.bars.strokeWidth = 0.5
        self.categoryAxis.categoryNames = ('1996', '1997', '1998', '1999', '2000')

        self.valueAxis.visible = 0
        self.categoryAxis.visible = 1
        self.categoryAxis.labels.fontName = "Helvetica-Oblique"
        self.categoryAxis.labels.fontSize = 6
        self.categoryAxis.labels.fillColor = _PCMYK_black
        self.categoryAxis.labels.dy = -self.labelOffset

        self.legend=Legend()
        self.legend.columnMaximum = 3
        self.legend.colorNamePairs = [(c0, "Fund"), (c1, "Index")]
        self.legend.x = self.width
        self.legend.y = self.lineLength
        self.legend.fontName = "Helvetica"
        self.legend.fontSize = 6
        self.legend.deltax = 0.1*cm
        self.legend.deltay = 0.4*cm
        self.legend.dxTextSpace = 5
        self.legend.dx = 5
        self.legend.dy = 5
        self.legend.alignment = 'right'
        self.legend.strokeColor = _PCMYK_black
        self.legend.strokeWidth = 0.5
        self.legend.boxAnchor = 'w'
        tx=0
        for f in range (0, len(self.legend.colorNamePairs)):
            tx=tx+_maxWidth(self.legend.colorNamePairs[f][1], self.legend.fontName, self.legend.fontSize)
            tx=tx+self.legend.deltax+self.legend.dxTextSpace+self.legend.dx

        self.leftPadding   = 5
        self.rightPadding  = 5
        self.topPadding    = 5
        self.bottomPadding = 5

    def _labelXY(self,label,x,y,width,height):
        'Compute x, y for a label'
        pm = height>=0 and 1 or -1
        self._collectedData.append((x + 0.5*width, y*pm, height))
        pm *= height
        label._pmv = pm
        return x+0.5*width, y+self.lineLength+(self.labelOffset*2)+self.barLabels.nudge, pm

    def draw(self):
        g = Group()
        if self.barLabelFormat == None:
            self.barLabelFormat = _nullFormatter
        self._collectedData=[]

        tl=[]
        for f in range (0, len(self.data)):
            for i in range (0,len(self.data[f])):
                tl.append(self.data[f][i])
        self.valueAxis.valueMax = _highest([tl])

        #if we have a background colour,do a background
        if self.background:
            w,h = self._getDrawingDimensions()
            g.add(Rect(-self.leftPadding,-(self.bottomPadding+self.labelFontSize+self.labelOffset),
                       x,y,
                       strokeColor=None,
                       strokeWidth=0,
                       fillColor=self.background))
        self.legend.x = self.x+self.width
        
        #the actual barcharts        
        g.add(barcharts.VerticalBarChart.draw(self))

        # the vertical lines
        for i in range(0, len(self._collectedData)):
            x,y, h=self._collectedData[i]
            g.add(Line(x, y+h, x, y+self.lineLength,
                       strokeColor = self.lineColor,
                       strokeWidth = self.lineStrokeWidth))
    
        g.add(self.legend)

        g.shift(self.leftPadding,self.bottomPadding+self.categoryAxis.labels.fontSize+self.labelOffset)
        return g

    def _getDrawingDimensions(self):
        #find width of grid
        tx=self.x+self.width
        #add padding (and offset)
        tx=tx+self.leftPadding+self.rightPadding+self.labelOffset
        #add in maximum width of text in legend
        tt=[]
        for f in range (0, len(self.legend.colorNamePairs)):
            tt.append(self.legend.colorNamePairs[f][1])
        tx=tx+_maxWidth(tt, self.legend.fontName, self.legend.fontSize)
        tx=tx+self.legend.deltax+self.legend.dxTextSpace+self.legend.dx
        #up to baseline...
        ty=self.bottomPadding+self.categoryAxis.labels.fontSize+self.labelOffset
        #add length of longest label (including percent signs, yadda, yadda, yadda)
        tl=[]
        for f in range (0, len(self.data)):
            for i in range (0,len(self.data[f])):
                if self.barLabelFormat is None:
                    tl.append(self.data[0][f])
                elif isinstance(self.barLabelFormat,str):
                    tl.append(self.barLabelFormat % self.data[0][f])
                elif iscallable(self.barLabelFormat):
                    tl.append(self.barLabelFormat(self.data[0][f]))
        ty=ty+self.lineLength+_maxWidth(tl, self.barLabels.fontName, self.barLabels.fontSize)
        ty=ty+self.topPadding
        #print (tx, ty)
        return (tx,ty)

    def demo(self, drawing=None):
        if not drawing:
            tx,ty=self._getDrawingDimensions()
            drawing = Drawing(tx, ty)
        drawing.add(self.draw())
        return drawing       


if __name__ == "__main__":
    d = VerticalBarChart()
    d.demo().save(fnRoot="d_verticalbarchart")



from unittest2.test.support import LoggingResult

import unittest2


class Test_TestSkipping(unittest2.TestCase):

    def test_skipping(self):
        class Foo(unittest2.TestCase):
            def test_skip_me(self):
                self.skipTest("skip")
        events = []
        result = LoggingResult(events)
        test = Foo("test_skip_me")
        test.run(result)
        self.assertEqual(events, ['startTest', 'addSkip', 'stopTest'])
        self.assertEqual(result.skipped, [(test, "skip")])

        # Try letting setUp skip the test now.
        class Foo(unittest2.TestCase):
            def setUp(self):
                self.skipTest("testing")
            def test_nothing(self): pass
        events = []
        result = LoggingResult(events)
        test = Foo("test_nothing")
        test.run(result)
        self.assertEqual(events, ['startTest', 'addSkip', 'stopTest'])
        self.assertEqual(result.skipped, [(test, "testing")])
        self.assertEqual(result.testsRun, 1)

    def test_skipping_decorators(self):
        op_table = ((unittest2.skipUnless, False, True),
                    (unittest2.skipIf, True, False))
        for deco, do_skip, dont_skip in op_table:
            class Foo(unittest2.TestCase):
                @deco(do_skip, "testing")
                def test_skip(self): 
                    pass

                @deco(dont_skip, "testing")
                def test_dont_skip(self): 
                    pass
            
            test_do_skip = Foo("test_skip")
            test_dont_skip = Foo("test_dont_skip")
            suite = unittest2.TestSuite([test_do_skip, test_dont_skip])
            events = []
            result = LoggingResult(events)
            suite.run(result)
            self.assertEqual(len(result.skipped), 1)
            expected = ['startTest', 'addSkip', 'stopTest',
                        'startTest', 'addSuccess', 'stopTest']
            self.assertEqual(events, expected)
            self.assertEqual(result.testsRun, 2)
            self.assertEqual(result.skipped, [(test_do_skip, "testing")])
            self.assertTrue(result.wasSuccessful())
        
    def test_skip_class(self):
        class Foo(unittest2.TestCase):
            def test_1(self):
                record.append(1)
        
        # was originally a class decorator...
        Foo = unittest2.skip("testing")(Foo)
        record = []
        result = unittest2.TestResult()
        test = Foo("test_1")
        suite = unittest2.TestSuite([test])
        suite.run(result)
        self.assertEqual(result.skipped, [(test, "testing")])
        self.assertEqual(record, [])

    def test_expected_failure(self):
        class Foo(unittest2.TestCase):
            @unittest2.expectedFailure
            def test_die(self):
                self.fail("help me!")
        events = []
        result = LoggingResult(events)
        test = Foo("test_die")
        test.run(result)
        self.assertEqual(events,
                         ['startTest', 'addExpectedFailure', 'stopTest'])
        self.assertEqual(result.expectedFailures[0][0], test)
        self.assertTrue(result.wasSuccessful())

    def test_unexpected_success(self):
        class Foo(unittest2.TestCase):
            @unittest2.expectedFailure
            def test_die(self):
                pass
        events = []
        result = LoggingResult(events)
        test = Foo("test_die")
        test.run(result)
        self.assertEqual(events,
                         ['startTest', 'addUnexpectedSuccess', 'stopTest'])
        self.assertFalse(result.failures)
        self.assertEqual(result.unexpectedSuccesses, [test])
        self.assertTrue(result.wasSuccessful())

    def test_skip_doesnt_run_setup(self):
        class Foo(unittest2.TestCase):
            wasSetUp = False
            wasTornDown = False
            def setUp(self):
                Foo.wasSetUp = True
            def tornDown(self):
                Foo.wasTornDown = True
            @unittest2.skip('testing')
            def test_1(self):
                pass
        
        result = unittest2.TestResult()
        test = Foo("test_1")
        suite = unittest2.TestSuite([test])
        suite.run(result)
        self.assertEqual(result.skipped, [(test, "testing")])
        self.assertFalse(Foo.wasSetUp)
        self.assertFalse(Foo.wasTornDown)
    
    def test_decorated_skip(self):
        def decorator(func):
            def inner(*a):
                return func(*a)
            return inner
        
        class Foo(unittest2.TestCase):
            @decorator
            @unittest2.skip('testing')
            def test_1(self):
                pass
        
        result = unittest2.TestResult()
        test = Foo("test_1")
        suite = unittest2.TestSuite([test])
        suite.run(result)
        self.assertEqual(result.skipped, [(test, "testing")])


if __name__ == '__main__':
    unittest2.main()



"""
pageCatcher demo script:
   reverse the pages in a pdf file.   
"""
from reportlab.pdfgen import canvas
inputFileName = "sample2.pdf"
outputFileName = "out2_reversed.pdf"
storageFile = "sample2.data"


try:
    from rlextra.pageCatcher.pageCatcher import storeForms, restoreForms
except ImportError:
    pass # running inside pageCatcher module?

def run(inputFileName=inputFileName, outputFileName=outputFileName, storageFile=storageFile):
    print("reversing", inputFileName, "into", outputFileName)
    storeForms(inputFileName, storageFile, all=1)
    print("stored all pages to", storageFile)
    canv = canvas.Canvas(outputFileName)
    formNames = restoreForms(storageFile, canv)

    # write out the pages in reversed order
    formNames.reverse()
    for f in formNames:
        canv.doForm(f)
        canv.showPage()
    canv.save()
    print("wrote", outputFileName)

if __name__=='__main__':
    run(inputFileName=inputFileName, outputFileName=outputFileName, storageFile=storageFile)

    



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""A library for reading and converting SVG.

This is a converter from SVG to RLG (ReportLab Graphics) drawings.
It converts mainly basic shapes, paths and simple text. The intended
usage is either as module within other projects:

    from svglib.svglib import svg2rlg
    drawing = svg2rlg("foo.svg")

or from the command-line where it is usable as an SVG to PDF converting
tool named sv2pdf (which should also handle SVG files compressed with
gzip and extension .svgz).
"""
from reportlab import xrange
import copy
import gzip
import itertools
import logging
import os, sys
import re
import base64
import tempfile
import subprocess
from collections import defaultdict, namedtuple
from functools import partial

from reportlab.pdfbase.pdfmetrics import registerFont, stringWidth
from reportlab.pdfbase.ttfonts import TTFError, TTFont
from reportlab.pdfgen.canvas import FILL_EVEN_ODD, FILL_NON_ZERO
from reportlab.pdfgen.pdfimages import PDFImage
from reportlab.graphics.shapes import (
    _CLOSEPATH, Circle, Drawing, Ellipse, Group, Image, Line, Path, PolyLine,
    Polygon, Rect, SolidShape, String,
)
from reportlab.lib import colors
from reportlab.lib.units import pica, toLength
try:
    from lxml import etree
except ImportError:
    import xml.dom.minidom 
    etree = None

from .utils import (
    bezier_arc_from_end_points, convert_quadratic_to_cubic_path,
    normalise_svg_path,
)

__version__ = '0.8.1'
__license__ = 'LGPL 3'
__author__ = 'Dinu Gherman'
__date__ = '2017-04-22'

XML_NS = 'http://www.w3.org/XML/1998/namespace'
STANDARD_FONT_NAMES = (
    'Times-Roman', 'Times-Italic', 'Times-Bold', 'Times-BoldItalic',
    'Helvetica', 'Helvetica-Oblique', 'Helvetica-Bold', 'Helvetica-BoldOblique',
    'Courier', 'Courier-Oblique', 'Courier-Bold', 'Courier-BoldOblique',
    'Symbol', 'ZapfDingbats',
)
DEFAULT_FONT_NAME = "Helvetica"
_registered_fonts = {}

logger = logging.getLogger(__name__)

Box = namedtuple('Box', ['x', 'y', 'width', 'height'])


def find_font(font_name, exact=False):
    """Return the font and a Boolean indicating if the match is exact."""
    if font_name in STANDARD_FONT_NAMES:
        return font_name, True
    elif font_name in _registered_fonts:
        return font_name, _registered_fonts[font_name]

    NOT_FOUND = (None, False)
    try:
        # Try first to register the font if it exists as ttf,
        # based on ReportLab font search.
        registerFont(TTFont(font_name, '%s.ttf' % font_name))
        _registered_fonts[font_name] = True
        return font_name, True
    except TTFError:
        # Try searching with Fontconfig
        try:
            pipe = subprocess.Popen(
                ['fc-match', '-s', '--format=%{file}\\n', font_name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            output = pipe.communicate()[0].decode(sys.getfilesystemencoding())
            font_path = output.split('\n')[0]
        except OSError:
            return NOT_FOUND
        try:
            registerFont(TTFont(font_name, font_path))
        except TTFError:
            return NOT_FOUND
        # Fontconfig may return a default font totally unrelated with font_name
        exact = font_name.lower() in os.path.basename(font_path).lower()
        _registered_fonts[font_name] = exact
        return font_name, exact



class ClippingPath(Path):
    def __init__(self, *args, **kwargs):
        copy_from = kwargs.pop('copy_from', None)
        Path.__init__(self, *args, **kwargs)
        if copy_from:
            self.__dict__.update(copy.deepcopy(copy_from.__dict__))
        self.isClipPath = 1

    def getProperties(self, *args, **kwargs):
        props = Path.getProperties(self, *args, **kwargs)
        if 'fillColor' in props:
            props['fillColor'] = None
        return props

class DeferredClippingPath(ClippingPath):
    def __init__(self,ref=None):
        Path.__init__(self)
        self._ref = ref

# Attribute converters (from SVG to RLG)

class AttributeConverter:
    "An abstract class to locate and convert attributes in a DOM instance."

    def parseMultiAttributes(self, line):
        """Try parsing compound attribute string.

        Return a dictionary with single attributes in 'line'.
        """

        attrs = line.split(';')
        attrs = [a.strip() for a in attrs]
        attrs = filter(lambda a:len(a)>0, attrs)

        new_attrs = {}
        for a in attrs:
            k, v = a.split(':')
            k, v = [s.strip() for s in (k, v)]
            new_attrs[k] = v

        return new_attrs

    def findAttr(self, svgNode, name):
        """Search an attribute with some name in some node or above.

        First the node is searched, then its style attribute, then
        the search continues in the node's parent node. If no such
        attribute is found, '' is returned.
        """

        # This needs also to lookup values like "url(#SomeName)"...

        attr_value = svgNode.getAttribute(name).strip()

        if attr_value and attr_value != "inherit":
            return attr_value
        elif svgNode.getAttribute("style"):
            d = self.parseMultiAttributes(svgNode.getAttribute("style"))
            if name in d:
                return d[name]
        if svgNode.getparent() is not None:
            return self.findAttr(svgNode.getparent(), name)

        return ''

    def getAllAttributes(self, svgNode):
        "Return a dictionary of all attributes of svgNode or those inherited by it."

        d = {}

        parent = svgNode.getparent()
        if parent and parent.nodeName == 'g':
            d.update(self.getAllAttributes(svgNode.getparent()))

        style = svgNode.getAttribute("style")
        if style:
            d = self.parseMultiAttributes(style)
            d.update(d)

        for key, value in svgNode.attrib.items():
            if key != "style":
                d[key] = value

        return d

    def id(self, svgAttr):
        "Return attribute as is."
        return svgAttr

    def convertTransform(self, svgAttr):
        """Parse transform attribute string.

        E.g. "scale(2) translate(10,20)"
             -> [("scale", 2), ("translate", (10,20))]
        """

        line = svgAttr.strip()

        ops = line[:]
        brackets = []
        indices = []
        for i, lin in enumerate(line):
            if lin in "()":
                brackets.append(i)
        for i in range(0, len(brackets), 2):
            bi, bj = brackets[i], brackets[i+1]
            subline = line[bi+1:bj]
            subline = subline.strip()
            subline = subline.replace(',', ' ')
            subline = re.sub("[ ]+", ',', subline)
            if ',' in subline:
                indices.append(tuple(float(num) for num in subline.split(',')))
            else:
                indices.append(float(subline))
            ops = ops[:bi] + ' '*(bj-bi+1) + ops[bj+1:]
        ops = ops.replace(',', ' ').split()

        if len(ops) != len(indices):
            logger.warn("Unable to parse transform expression '%s'" % svgAttr)
            return []

        result = []
        for i, op in enumerate(ops):
            result.append((op, indices[i]))

        return result


class Svg2RlgAttributeConverter(AttributeConverter):
    "A concrete SVG to RLG attribute converter."

    def __init__(self,color_converter=None):
        self.color_converter = color_converter or self.identity_color_converter

    @staticmethod
    def identity_color_converter(c):
        return c

    def convertLength(self, svgAttr, percentOf=100, em_base=12):
        "Convert length to points."

        text = svgAttr
        if not text:
            return 0.0
        if ' ' in text.replace(',', ' ').strip():
            logger.debug("Only getting first value of %s" % text)
            text = text.replace(',', ' ').split()[0]

        if text.endswith('%'):
            logger.debug("Fiddling length unit: %")
            return float(text[:-1]) / 100 * percentOf
        elif text.endswith("pc"):
            return float(text[:-2]) * pica
        elif text.endswith("pt"):
            return float(text[:-2]) * 1.25
        elif text.endswith("em"):
            return float(text[:-2]) * em_base
        elif text.endswith("px"):
            return float(text[:-2])

        for unit in ("ex",):
            if unit in text:
                logger.warn("Ignoring unit: %s" % unit)
                text = text.replace(unit, '')

        text = text.strip()
        length = toLength(text)

        return length

    _re_spaces = re.compile(r'\s+')
    @classmethod
    def split_attr_list(cls,svgAttr):
        return cls._re_spaces.split(svgAttr.strip().replace(',', ' '))

    def convertLengthList(self, svgAttr):
        "Convert a list of lengths."
        return [self.convertLength(a) for a in self.split_attr_list(svgAttr)]

    def convertOpacity(self, svgAttr):
        return float(svgAttr)

    def convertFillRule(self, svgAttr):
        return {
            'nonzero': FILL_NON_ZERO,
            'evenodd': FILL_EVEN_ODD,
        }.get(svgAttr, '')

    def convertColor(self, svgAttr):
        "Convert string to a RL color object."

        # fix it: most likely all "web colors" are allowed
        predefined = "aqua black blue fuchsia gray green lime maroon navy "
        predefined = predefined + "olive orange purple red silver teal white yellow "
        predefined = predefined + "lawngreen indianred aquamarine lightgreen brown"

        # This needs also to lookup values like "url(#SomeName)"...

        text = svgAttr
        if not text or text == "none":
            return None

        if text in predefined.split():
            return self.color_converter(getattr(colors, text))
        elif text == "currentColor":
            return "currentColor"
        elif len(text) == 7 and text[0] == '#':
            return self.color_converter(colors.HexColor(text))
        elif len(text) == 4 and text[0] == '#':
            return self.color_converter(colors.HexColor('#' + 2*text[1] + 2*text[2] + 2*text[3]))
        elif text.startswith('rgb') and '%' not in text:
            t = text[3:].strip('()')
            tup = [h[2:] for h in [hex(int(num)) for num in t.split(',')]]
            tup = [(2 - len(h)) * '0' + h for h in tup]
            col = "#%s%s%s" % tuple(tup)
            return self.color_converter(colors.HexColor(col))
        elif text.startswith('rgb') and '%' in text:
            t = text[3:].replace('%', '').strip('()')
            tup = (int(val)/100.0 for val in t.split(','))
            return self.color_converter(colors.Color(*tup))

        logger.warn("Can't handle color: %s" % text)

        return None

    def convertLineJoin(self, svgAttr):
        return {"miter":0, "round":1, "bevel":2}[svgAttr]

    def convertLineCap(self, svgAttr):
        return {"butt":0, "round":1, "square":2}[svgAttr]

    def convertDashArray(self, svgAttr):
        strokeDashArray = self.convertLengthList(svgAttr)
        return strokeDashArray

    def convertDashOffset(self, svgAttr):
        strokeDashOffset = self.convertLength(svgAttr)
        return strokeDashOffset

    def convertFontFamily(self, svgAttr):
        # very hackish
        font_mapping = {
            "sans-serif": "Helvetica",
            "serif": "Times-Roman",
            "times": "Times-Roman",
            "monospace": "Courier",
            }

        font_names = [
            font_mapping.get(font_name.lower(), font_name)
            for font_name in self.split_attr_list(svgAttr)
        ]
        non_exact_matches = []
        for font_name in font_names:
            font_name, exact = find_font(font_name)
            if exact:
                return font_name
            elif font_name:
                non_exact_matches.append(font_name)
        if non_exact_matches:
            return non_exact_matches[0]
        else:
            logger.warn("Unable to find a suitable font for 'font-family:%s'" % svgAttr)
            return DEFAULT_FONT_NAME

        return font_name


class NodeCompat(object):
    """A wrapper to make minidom and lxml look alike"""

    def __init__(self, node):
        self.node = node

    if etree:
        def getAttribute(self, name):
            return self.node.attrib.get(name, '')

        def getAttributeNS(self, uri,name):
            return self.getAttribute("{%s}%s" % (uri,name))

        def getchildren(self,allNodes=False):
            return [NodeCompat(c) for c in self.node.getchildren()]

        @property
        def nodeName(self):
            """Return lxml node name without the namespace prefix."""
            try:
                return self.tag.split('}')[-1]
            except AttributeError:
                pass

        @property
        def is_element(self):
            return self.nodeName!='text'

        @property
        def is_text(self):
            return self.nodeName=='text'

        def getparent(self):
            p = self.node.getparent()
            if p is not None: return NodeCompat(p)

        def deepClone(self):
            return NodeCompat(copy.deepcopy(self.node))

        def appendChild(self,c):
            self.node.append(c.node)

    else:
        def getAttribute(self, name):
            # forward call to wrapped node
            return self.node.getAttribute(name) if self.node.nodeType==self.node.ELEMENT_NODE else ''

        @property
        def attrib(self):
            return self.node.attributes

        @property
        def is_text(self):
            return self.node.nodeType in (self.node.TEXT_NODE,self.node.CDATA_SECTION_NODE)

        @property
        def text(self):
            if self.node.nodeType in (self.node.TEXT_NODE,self.node.CDATA_SECTION_NODE):
                return self.data
            else:
                return ''.join(c.data for c in self.node.childNodes if c.nodeType in (self.node.TEXT_NODE,self.node.CDATA_SECTION_NODE))

        @property
        def is_element(self):
            return self.node.nodeType==self.node.ELEMENT_NODE

        def getparent(self):
            p = self.node.parentNode
            if p is not None: return NodeCompat(p)

        def getchildren(self,allNodes=False):
            return [NodeCompat(c) for c in self.node.childNodes if allNodes or c.nodeType==self.node.ELEMENT_NODE]

        def deepClone(self):
            return NodeCompat(self.node.cloneNode(True))

        def getAttributeNS(self,uri,name):
            return self.node.getAttributeNS(uri,name)

        def appendChild(self,c):
            self.node.appendChild(c.node)

        @property
        def attrib(self):
            return self.node.attributes

    def __getattr__(self, name):
        # forward attribute access to wrapped node
        return getattr(self.node, name)

class NodeTracker:
    """An object wrapper keeping track of arguments to certain method calls.

    Instances wrap an object and store all arguments to one special
    method, getAttribute(name), in a list of unique elements, usedAttrs.
    """
    def __init__(self, node):
        if not isinstance(node,NodeCompat):
            node = NodeCompat(node)
        self.node = node
        self.usedAttrs = []

    def getAttribute(self, name):
        # add argument to the history, if not already present
        if name not in self.usedAttrs:
            self.usedAttrs.append(name)
        # forward call to wrapped object
        return  self.node.getAttribute(name)

    if not etree:
        def getAttributeNS(self, uri, name):
            # add argument to the history, if not already present
            try:
                attr = self._attrsNS[(uri,name)]
            except KeyError:
                return  self.node.getAttributeNS(uri,name)
            name = attr.name
            if name not in self.usedAttrs:
                self.usedAttrs.append(name)
            return attr.value

    def __getattr__(self, name):
        # forward attribute access to wrapped object
        return getattr(self.node, name)

def get_shape_from_group(G):
    #for clip path handling
    for e in G.contents:
        if isinstance(e, Group):
            return get_shape_from_group(e)
        elif isinstance(e, SolidShape):
            return e

def copy_properties(src, dst):
    for p, v in src.getProperties().items():
        try:
            setattr(dst, p, v)
        except AttributeError:
            pass

def convert_shape_to_clipping_path(shape):
    if isinstance(shape, Path):
        return ClippingPath(copy_from=shape)
    elif isinstance(shape, Rect):
        # It is possible to use a rect as a clipping path in an svg, so we
        # need to convert it to a path for rlg.
        x1, y1, x2, y2 = shape.getBounds()
        p = ClippingPath()
        p.moveTo(x1, y1)
        p.lineTo(x2, y1)
        p.lineTo(x2, y2)
        p.lineTo(x1, y2)
        p.closePath()
        # Copy the styles from the rect to the clipping path.
        copy_properties(shape, p)
        return p
    elif shape:
        logging.error("Unsupported shape type %s for clipping" % shape.__class__.__name__)

### the main meat ###

class SvgRenderer:
    """Renderer that renders an SVG file on a ReportLab Drawing instance.

    This is the base class for walking over an SVG DOM document and
    transforming it into a ReportLab Drawing instance.
    """

    def __init__(self, path=None,color_converter=None):
        self.attrConverter = Svg2RlgAttributeConverter(color_converter=color_converter)
        self.shape_converter = Svg2RlgShapeConverter(path,self.attrConverter)
        self.handled_shapes = self.shape_converter.get_handled_shapes()
        self.definitions = {}
        self.waiting_use_nodes = defaultdict(list)
        self.wanted_clip_paths = defaultdict(list)
        self.box = Box(x=0, y=0, width=0, height=0)

    def render(self, svg_node):
        main_group = self.renderNode(svg_node)
        for xlink in self.waiting_use_nodes.keys():
            logger.debug("Ignoring unavailable object with ID '%s'." % xlink)

        main_group.scale(1, -1)
        main_group.translate(0 - self.box.x, -self.box.height - self.box.y)
        drawing = Drawing(self.box.width, self.box.height)
        drawing.add(main_group)
        return drawing

    def renderNode(self, node, parent=None):
        n = NodeTracker(node)
        nid = n.getAttribute("id")
        ignored = False
        item = None
        name = n.nodeName

        clipping = self.get_clippath(n)
        if name == "svg":
            if n.getAttributeNS(XML_NS,"space") == 'preserve':
                self.shape_converter.preserve_space = True
            item = self.renderSvg(n)
            return item
        elif name == "defs":
            item = self.renderG(n)
        elif name == 'a':
            item = self.renderA(n)
            parent.add(item)
        elif name == 'g':
            display = n.getAttribute("display")
            item = self.renderG(n, clipping=clipping)
            if display != "none":
                parent.add(item)
        elif name == "symbol":
            item = self.renderSymbol(n)
            parent.add(item)
        elif name == "use":
            item = self.renderUse(n, clipping=clipping)
            parent.add(item)
        elif name == "clipPath":
            item = self.renderG(n)
        elif name in self.handled_shapes:
            display = n.getAttribute("display")
            item = self.shape_converter.convertShape(name, n, clipping, self)
            if item and display != "none":
                parent.add(item)
        else:
            ignored = True
            logger.debug("Ignoring node: %s" % name)

        if not ignored:
            if nid and item:
                self.definitions[nid] = node
                if nid in self.wanted_clip_paths:
                    p = convert_shape_to_clipping_path(get_shape_from_group(item))
                    for g in self.wanted_clip_paths.pop(nid,[]):
                        #backfill the DeferredClippingPaths in group g
                        for i, c in enumerate(g.contents):
                            if getattr(c,'_ref',None)==nid:
                                g.contents[i] = ClippingPath(copy_from=p)
            if nid in self.waiting_use_nodes.keys():
                to_render = self.waiting_use_nodes.pop(nid)
                for use_node, group in to_render:
                    self.renderUse(use_node, group=group)
            if getattr(n,'_node_rendered',True):
                self.print_unused_attributes(node, n)

    def get_shape_from_node(self,node):
        for c in node.getchildren():
            if c.nodeName == 'path':
                g = self.shape_converter.convertShape('path', NodeTracker(c))
                return g.contents[-1]
            elif c.nodeName == 'use':
                g = self.renderUse(NodeTracker(c))
                return get_shape_from_group(g)
            elif  c.nodeName == 'rect':
                return self.shape_converter.convertRect(NodeTracker(c))
            else:
                return get_shape_from_node(c)

    def get_clippath(self, node):
        """
        Return the clipping Path object referenced by the node 'clip-path'
        attribute, if any.
        """
        clip_path = node.getAttribute('clip-path')
        if clip_path:
            m = re.match(r'url\(#([^\)]*)\)', clip_path)
            if m:
                ref = m.groups()[0]
                if ref in self.definitions:
                    return convert_shape_to_clipping_path(self.get_shape_from_node(self.definitions[ref]))
                else:
                    return DeferredClippingPath(ref=ref)

    def print_unused_attributes(self, node, n):
        if logger.level > logging.DEBUG:
            return
        all_attrs = self.attrConverter.getAllAttributes(node)
        unused_attrs = ['%s=%r' % (attr[0],attr[1]) for attr in all_attrs.items() if attr[0] not in n.usedAttrs]
        if unused_attrs:
            logger.debug("Unused attrs: %s %s" % (n.nodeName, unused_attrs))

    def renderTitle_(self, node):
        # Main SVG title attr. could be used in the PDF document info field.
        pass

    def renderDesc_(self, node):
        # Main SVG desc. attr. could be used in the PDF document info field.
        pass

    def renderSvg(self, node):
        getAttr = node.getAttribute
        width, height = map(getAttr, ("width", "height"))
        width, height = map(self.attrConverter.convertLength, (width, height))
        viewBox = getAttr("viewBox")
        if viewBox:
            viewBox = self.attrConverter.convertLengthList(viewBox)
            self.box = Box(*viewBox)
        else:
            self.box = Box(0, 0, width, height)
        group = Group()
        for child in node.getchildren():
            self.renderNode(child, group)
        return group

    def add_clipping_to_group(self,clipping,group):
        group.add(clipping)
        if isinstance(clipping,DeferredClippingPath):
            self.wanted_clip_paths[clipping._ref].append(group)

    def renderG(self, node, clipping=None, display=1):
        getAttr = node.getAttribute
        id, transform = map(getAttr, ("id", "transform"))
        gr = Group()
        if clipping:
            self.add_clipping_to_group(clipping, gr)
        for child in node.getchildren():
            item = self.renderNode(child, parent=gr)
            if item and display:
                gr.add(item)

        if transform:
            self.shape_converter.applyTransformOnGroup(transform, gr)

        return gr

    def renderSymbol(self, node):
        return self.renderG(node, display=0)

    def renderA(self, node):
        # currently nothing but a group...
        # there is no linking info stored in shapes, maybe a group should?
        return self.renderG(node)

    def renderUse(self, node, group=None, clipping=None):
        xlink_href = node.getAttributeNS('http://www.w3.org/1999/xlink','href')
        if not xlink_href:
            node._node_rendered = False
            return
        if group is None:
            group = Group()
        uid = xlink_href[1:]
        if uid not in self.definitions:
            # The missing definition should appear later in the file
            self.waiting_use_nodes[uid].append((node, group))
            node._node_rendered = False
            return group

        if clipping:
            self.add_clipping_to_group(clipping, group)
        if len(node.getchildren()) == 0:
            # Append a copy of the referenced node as the <use> child (if not already done)
            node.appendChild(self.definitions[uid].deepClone())
        self.renderNode(node.getchildren()[-1], parent=group)
        getAttr = node.getAttribute
        x, y, width, height, transform = [getAttr(a) for a in  ("x", "y", "width", "height", "transform")]
        if x or y:
            transform += " translate(%s, %s)" % (x or '0', y or '0')
        if transform:
            self.shape_converter.applyTransformOnGroup(transform, group)
        node._node_rendered = True
        return group


class SvgShapeConverter:
    """An abstract SVG shape converter.

    Implement subclasses with methods named 'renderX(node)', where
    'X' should be the capitalised name of an SVG node element for
    shapes, like 'Rect', 'Circle', 'Line', etc.

    Each of these methods should return a shape object appropriate
    for the target format.
    """
    def __init__(self, path, attrConverter=None):
        self.attrConverter = attrConverter or Svg2RlgAttributeConverter()
        self.svg_source_file = path
        self.preserve_space = False

    @classmethod
    def get_handled_shapes(cls):
        """Dynamically determine a list of handled shape elements based on
           convert<shape> method existence.
        """
        return [key[7:].lower() for key in dir(cls) if key.startswith('convert')]


class Svg2RlgShapeConverter(SvgShapeConverter):
    """Converter from SVG shapes to RLG (ReportLab Graphics) shapes."""

    def convertShape(self, name, node, clipping=None, renderer=None):
        method_name = "convert%s" % name.capitalize()
        shape = getattr(self, method_name)(node)
        if not shape:
            return
        if name not in ('path', 'polyline', 'text'):
            # Only apply style where the convert method did not apply it.
            self.applyStyleOnShape(shape, node)
        transform = node.getAttribute("transform")
        if not (transform or clipping):
            return shape
        else:
            group = Group()
            if transform:
                self.applyTransformOnGroup(transform, group)
            if clipping:
                renderer.add_clipping_to_group(clipping, group)
            group.add(shape)
            return group

    def convertLine(self, node):
        getAttr = node.getAttribute
        x1, y1, x2, y2 = map(getAttr, ("x1", "y1", "x2", "y2"))
        x1, y1, x2, y2 = map(self.attrConverter.convertLength, (x1, y1, x2, y2))
        shape = Line(x1, y1, x2, y2)

        return shape

    def convertRect(self, node):
        getAttr = node.getAttribute
        x, y, width, height = map(getAttr, ('x', 'y', "width", "height"))
        x, y, width, height = map(self.attrConverter.convertLength, (x, y, width, height))
        rx, ry = map(getAttr, ("rx", "ry"))
        rx, ry = map(self.attrConverter.convertLength, (rx, ry))
        shape = Rect(x, y, width, height, rx=rx, ry=ry)

        return shape

    def convertCircle(self, node):
        # not rendered if r == 0, error if r < 0.
        getAttr = node.getAttribute
        cx, cy, r = map(getAttr, ("cx", "cy", 'r'))
        cx, cy, r = map(self.attrConverter.convertLength, (cx, cy, r))
        shape = Circle(cx, cy, r)

        return shape

    def convertEllipse(self, node):
        getAttr = node.getAttribute
        cx, cy, rx, ry = map(getAttr, ("cx", "cy", "rx", "ry"))
        cx, cy, rx, ry = map(self.attrConverter.convertLength, (cx, cy, rx, ry))
        width, height = rx, ry
        shape = Ellipse(cx, cy, width, height)

        return shape

    def convertPolyline(self, node):
        getAttr = node.getAttribute
        points = getAttr("points")
        points = points.replace(',', ' ')
        points = points.split()
        points = list(map(self.attrConverter.convertLength, points))
        if len(points) % 2 != 0 or len(points) == 0:
            # Odd number of coordinates or no coordinates, invalid polyline
            return None

        polyline = PolyLine(points)
        self.applyStyleOnShape(polyline, node)
        has_fill = self.attrConverter.findAttr(node, 'fill') not in ('', 'none')

        if has_fill:
            # ReportLab doesn't fill polylines, so we are creating a polygon
            # polygon copy of the polyline, but without stroke.
            group = Group()
            polygon = Polygon(points)
            self.applyStyleOnShape(polygon, node)
            polygon.strokeColor = None
            group.add(polygon)
            group.add(polyline)
            return group

        return polyline

    def convertPolygon(self, node):
        getAttr = node.getAttribute
        points = getAttr("points")
        points = points.replace(',', ' ')
        points = points.split()
        points = list(map(self.attrConverter.convertLength, points))
        if len(points) % 2 != 0 or len(points) == 0:
            # Odd number of coordinates or no coordinates, invalid polygon
            return None
        shape = Polygon(points)

        return shape

    def clean_text(self, text, preserve_space):
        """Text cleaning as per https://www.w3.org/TR/SVG/text.html#WhiteSpace
        """
        if text is None:
            return
        if preserve_space:
            text = text.replace('\r\n', ' ').replace('\n', ' ').replace('\t', ' ')
        else:
            text = text.replace('\r\n', '').replace('\n', '').replace('\t', ' ')
            text = text.strip()
            while ('  ' in text):
                text = text.replace('  ', ' ')
        return text

    def convertText(self, node):
        attrConv = self.attrConverter
        xml_space = node.getAttributeNS(XML_NS,"space")
        if xml_space:
            preserve_space = xml_space == 'preserve'
        else:
            preserve_space = self.preserve_space

        gr = Group()

        frag_lengths = []

        dx0, dy0 = 0, 0
        x1, y1 = 0, 0
        ff = attrConv.findAttr(node, "font-family") or DEFAULT_FONT_NAME
        ff = attrConv.convertFontFamily(ff)
        fs = attrConv.findAttr(node, "font-size") or "12"
        fs = attrConv.convertLength(fs)
        convertLength = partial(attrConv.convertLength, em_base=fs)
        x, y = map(node.getAttribute, ('x', 'y'))
        x, y = map(convertLength, (x, y))
        for c in itertools.chain([node], node.getchildren(allNodes=True)):
            has_x, has_y = False, False
            dx, dy = 0, 0
            baseLineShift = 0
            if c.is_text:
                text = self.clean_text(c.text, preserve_space)
                if not text:
                    continue
            elif c.nodeName == 'tspan':
                text = self.clean_text(c.text, preserve_space)
                if not text:
                    continue
                x1, y1, dx, dy = [c.getAttribute(name) for name in ("x", "y", "dx", "dy")]
                has_x, has_y = (x1 != '', y1 != '')
                x1, y1, dx, dy = map(convertLength, (x1, y1, dx, dy))
                dx0 += dx
                dy0 += dy
                baseLineShift = c.getAttribute("baseline-shift") or ''
                if baseLineShift in ("sub", "super", "baseline"):
                    baseLineShift = {"sub":-fs/2, "super":fs/2, "baseline":0}[baseLineShift]
                else:
                    baseLineShift = convertLength(baseLineShift, fs)
            else:
                continue

            frag_lengths.append(stringWidth(text, ff, fs))
            new_x = (x1 + dx) if has_x else (x + dx0 + sum(frag_lengths[:-1]))
            new_y = (y1 + dy) if has_y else (y + dy0)
            shape = String(new_x, -(new_y - baseLineShift), text)
            self.applyStyleOnShape(shape, node)
            if c.nodeName == 'tspan':
                self.applyStyleOnShape(shape, c)

            gr.add(shape)

        gr.scale(1, -1)

        return gr

    def convertPath(self, node):
        d = node.getAttribute('d')
        if not d:
            return None
        normPath = normalise_svg_path(d)
        path = Path(autoclose='svg')
        points = path.points
        # Track subpaths needing to be closed later
        subpath_start = []
        lastop = ''

        for i in xrange(0, len(normPath), 2):
            op, nums = normPath[i:i+2]

            # moveto absolute
            if op == 'M':
                path.moveTo(*nums)
                subpath_start = points[-2:]
            # lineto absolute
            elif op == 'L':
                path.lineTo(*nums)

            # moveto relative
            elif op == 'm':
                if len(points) >= 2:
                    if lastop in ('Z', 'z'):
                        starting_point = subpath_start
                    else:
                        starting_point = points[-2:]
                    xn, yn = starting_point[0] + nums[0], starting_point[1] + nums[1]
                    path.moveTo(xn, yn)
                else:
                    path.moveTo(*nums)
                subpath_start = points[-2:]
            # lineto relative
            elif op == 'l':
                xn, yn = points[-2] + nums[0], points[-1] + nums[1]
                path.lineTo(xn, yn)

            # horizontal/vertical line absolute
            elif op == 'H':
                path.lineTo(nums[0], points[-1])
            elif op == 'V':
                path.lineTo(points[-2], nums[0])

            # horizontal/vertical line relative
            elif op == 'h':
                path.lineTo(points[-2] + nums[0], points[-1])
            elif op == 'v':
                path.lineTo(points[-2], points[-1] + nums[0])

            # cubic bezier, absolute
            elif op == 'C':
                path.curveTo(*nums)
            elif op == 'S':
                x2, y2, xn, yn = nums
                if len(points) < 4 or lastop not in {'c', 'C', 's', 'S'}:
                    xp, yp, x0, y0 = points[-2:] * 2
                else:
                    xp, yp, x0, y0 = points[-4:]
                xi, yi = x0 + (x0 - xp), y0 + (y0 - yp)
                path.curveTo(xi, yi, x2, y2, xn, yn)

            # cubic bezier, relative
            elif op == 'c':
                xp, yp = points[-2:]
                x1, y1, x2, y2, xn, yn = nums
                path.curveTo(xp + x1, yp + y1, xp + x2, yp + y2, xp + xn, yp + yn)
            elif op == 's':
                x2, y2, xn, yn = nums
                if len(points) < 4 or lastop not in {'c', 'C', 's', 'S'}:
                    xp, yp, x0, y0 = points[-2:] * 2
                else:
                    xp, yp, x0, y0 = points[-4:]
                xi, yi = x0 + (x0 - xp), y0 + (y0 - yp)
                path.curveTo(xi, yi, x0 + x2, y0 + y2, x0 + xn, y0 + yn)

            # quadratic bezier, absolute
            elif op == 'Q':
                x0, y0 = points[-2:]
                x1, y1, xn, yn = nums
                (x0, y0), (x1, y1), (x2, y2), (xn, yn) = \
                    convert_quadratic_to_cubic_path((x0, y0), (x1, y1), (xn, yn))
                path.curveTo(x1, y1, x2, y2, xn, yn)
            elif op == 'T':
                if len(points) < 4:
                    xp, yp, x0, y0 = points[-2:] * 2
                else:
                    xp, yp, x0, y0 = points[-4:]
                xi, yi = x0 + (x0 - xp), y0 + (y0 - yp)
                xn, yn = nums
                (x0, y0), (x1, y1), (x2, y2), (xn, yn) = \
                    convert_quadratic_to_cubic_path((x0, y0), (xi, yi), (xn, yn))
                path.curveTo(x1, y1, x2, y2, xn, yn)

            # quadratic bezier, relative
            elif op == 'q':
                x0, y0 = points[-2:]
                x1, y1, xn, yn = nums
                x1, y1, xn, yn = x0 + x1, y0 + y1, x0 + xn, y0 + yn
                (x0, y0), (x1, y1), (x2, y2), (xn, yn) = \
                    convert_quadratic_to_cubic_path((x0, y0), (x1, y1), (xn, yn))
                path.curveTo(x1, y1, x2, y2, xn, yn)
            elif op == 't':
                if len(points) < 4:
                    xp, yp, x0, y0 = points[-2:] * 2
                else:
                    xp, yp, x0, y0 = points[-4:]
                x0, y0 = points[-2:]
                xn, yn = nums
                xn, yn = x0 + xn, y0 + yn
                xi, yi = x0 + (x0 - xp), y0 + (y0 - yp)
                (x0, y0), (x1, y1), (x2, y2), (xn, yn) = \
                    convert_quadratic_to_cubic_path((x0, y0), (xi, yi), (xn, yn))
                path.curveTo(x1, y1, x2, y2, xn, yn)

            # elliptical arc
            elif op in ('A', 'a'):
                rx, ry, phi, fA, fS, x2, y2 = nums
                x1, y1 = points[-2:]
                if op == 'a':
                    x2 += x1
                    y2 += y1
                if abs(rx) <= 1e-10 or abs(ry) <= 1e-10:
                    path.lineTo(x2, y2)
                else:
                    bp = bezier_arc_from_end_points(x1, y1, rx, ry, phi, fA, fS, x2, y2)
                    for _, _, x1, y1, x2, y2, xn, yn in bp:
                        path.curveTo(x1, y1, x2, y2, xn, yn)

            # close path
            elif op in ('Z', 'z'):
                path.closePath()

            else:
                logger.debug("Suspicious path operator: %s" % op)
            lastop = op

        gr = Group()
        self.applyStyleOnShape(path, node)

        gr.add(path)
        return gr

    def convertImage(self, node):
        logger.warn("Adding box instead of image.")
        getAttr = node.getAttribute
        x, y, width, height = map(getAttr, ('x', 'y', "width", "height"))
        x, y, width, height = map(self.attrConverter.convertLength, (x, y, width, height))
        xlink_href = node.getAttributeNS('http://www.w3.org/1999/xlink','href')

        magic = "data:image/jpeg;base64"
        if xlink_href[:len(magic)] == magic:
            pat = "data:image/(\w+?);base64"
            ext = re.match(pat, magic).groups()[0]
            jpeg_data = base64.decodestring(xlink_href[len(magic):].encode('ascii'))
            _, path = tempfile.mkstemp(suffix='.%s' % ext)
            with open(path, 'wb') as fh:
                fh.write(jpeg_data)
            img = Image(int(x), int(y+height), int(width), int(-height), path)
            # this needs to be removed later, not here...
            # if exists(path): os.remove(path)
        else:
            xlink_href = os.path.join(os.path.dirname(self.svg_source_file), xlink_href)
            img = Image(int(x), int(y+height), int(width), int(-height), xlink_href)
            try:
                # this will catch invalid image
                PDFImage(xlink_href, 0, 0)
            except IOError:
                logger.error("Unable to read the image %s. Skipping..." % img.path)
                return None
        return img

    def applyTransformOnGroup(self, transform, group):
        """Apply an SVG transformation to a RL Group shape.

        The transformation is the value of an SVG transform attribute
        like transform="scale(1, -1) translate(10, 30)".

        rotate(<angle> [<cx> <cy>]) is equivalent to:
          translate(<cx> <cy>) rotate(<angle>) translate(-<cx> -<cy>)
        """

        tr = self.attrConverter.convertTransform(transform)
        for op, values in tr:
            if op == "scale":
                if not isinstance(values, tuple):
                    values = (values, values)
                group.scale(*values)
            elif op == "translate":
                if isinstance(values, (int, float)):
                    # From the SVG spec: If <ty> is not provided, it is assumed to be zero.
                    values = values, 0
                group.translate(*values)
            elif op == "rotate":
                if not isinstance(values, tuple) or len(values) == 1:
                    group.rotate(values)
                elif len(values) == 3:
                    angle, cx, cy = values
                    group.translate(cx, cy)
                    group.rotate(angle)
                    group.translate(-cx, -cy)
            elif op == "skewX":
                group.skew(values, 0)
            elif op == "skewY":
                group.skew(0, values)
            elif op == "matrix":
                group.transform = values
            else:
                logger.debug("Ignoring transform: %s %s" % (op, values))

    def applyStyleOnShape(self, shape, node, only_explicit=False):
        """
        Apply styles from an SVG element to an RLG shape.
        If only_explicit is True, only attributes really present are applied.
        """

        # RLG-specific: all RLG shapes
        "Apply style attributes of a sequence of nodes to an RL shape."

        # tuple format: (svgAttr, rlgAttr, converter, default)
        mappingN = (
            ("fill", "fillColor", "convertColor", "black"),
            ("fill-opacity", "fillOpacity", "convertOpacity", 1),
            ("fill-rule", "fillMode", "convertFillRule", "nonzero"),
            ("stroke", "strokeColor", "convertColor", "none"),
            ("stroke-width", "strokeWidth", "convertLength", "1"),
            ("stroke-opacity", "strokeOpacity", "convertOpacity", 1),
            ("stroke-linejoin", "strokeLineJoin", "convertLineJoin", "0"),
            ("stroke-linecap", "strokeLineCap", "convertLineCap", "0"),
            ("stroke-dasharray", "strokeDashArray", "convertDashArray", "none"),
        )
        mappingF = (
            ("font-family", "fontName", "convertFontFamily", DEFAULT_FONT_NAME),
            ("font-size", "fontSize", "convertLength", "12"),
            ("text-anchor", "textAnchor", "id", "start"),
        )

        if shape.__class__ == Group:
            # Recursively apply style on Group subelements
            for subshape in shape.contents:
                self.applyStyleOnShape(subshape, node, only_explicit=only_explicit)
            return

        ac = self.attrConverter
        for mapping in (mappingN, mappingF):
            if shape.__class__ != String and mapping == mappingF:
                continue
            for (svgAttrName, rlgAttr, func, default) in mapping:
                svgAttrValue = ac.findAttr(node, svgAttrName)
                if svgAttrValue == '':
                    if only_explicit:
                        continue
                    else:
                        svgAttrValue = default
                if svgAttrValue == "currentColor":
                    svgAttrValue = ac.findAttr(node.getparent(), "color") or default
                try:
                    meth = getattr(ac, func)
                    setattr(shape, rlgAttr, meth(svgAttrValue))
                except (AttributeError, KeyError, ValueError):
                    pass
        if getattr(shape, 'fillOpacity', None) is not None and shape.fillColor:
            shape.fillColor.alpha = shape.fillOpacity


def svg2rlg(path,**kwds):
    "Convert an SVG file to an RLG Drawing object."

    # unzip .svgz file into .svg
    unzipped = False
    if isinstance(path, str) and os.path.splitext(path)[1].lower() == ".svgz":
        data = gzip.GzipFile(path, "rb").read()
        open(path[:-1], 'w').write(data)
        path = path[:-1]
        unzipped = True

    # load SVG file
    if etree:
        parser = etree.XMLParser(remove_comments=True, recover=True)
        try:
            doc = etree.parse(path, parser=parser)
            svg = doc.getroot()
        except Exception as exc:
            logger.error("Failed to load input file, %r! (%s)" % (path,exc))
            return
    else:
        try:
            doc = xml.dom.minidom.parse(path)
            svg = doc.documentElement
        except Exception as exc:
            print("!!!!! Failed to load input file, %r! (%s)!" % (path,exc))
            return

    # convert to a RLG drawing
    svgRenderer = SvgRenderer(path,**kwds)
    drawing = svgRenderer.render(svg)

    # remove unzipped .svgz file (.svg)
    if unzipped:
        os.remove(path)

    return drawing


def main():
    import glob, traceback, sys
    argv = sys.argv[1:]
    outDir = os.getcwd()
    formats = ['pdf']
    verbose = logging.CRITICAL
    handling = False
    for arg in argv:
        if arg.startswith('--outdir='):
            outDir = arg[9:]
            if not os.path.isabs(outDir):
                outDir = os.path.abspath(outDir)
            continue
        elif arg.startswith('--formats='):
            formats = [_.strip() for _ in arg[10:].split(',')]
            continue
        elif arg.startswith('--verbose='):
            verbose = max(int(arg[10:]),logging.CRITICAL)
            continue
        elif arg=='--verbose':
            verbose = logging.DEBUG
            continue
        elif arg.startswith('--'):
            raise ValueError('unknown option %r' % arg)
        if verbose!=None:
            if not handling:
                logger.addHandler(logging.StreamHandler(sys.stdout))
                handling = True
            logger.setLevel(verbose)
        for fn in glob.glob(arg):
            try:
                d = svg2rlg(fn)
                if d:
                    fnRoot=os.path.splitext(os.path.basename(fn))[0]
                    d.save(formats=formats,verbose=verbose,fnRoot=fnRoot,outDir=outDir)
            except:
                traceback.print_exc()

if __name__=='__main__':
    main()
